// ppio.c

//;-------------------------------------------------------------------------
//; Copyright (c) 2002 Steven Freyder, David Helland, and Bruce D. Lightner.
//; Use subject to licensing restrictions described in "PICOWEB SERVER
//; SOFTWARE LICENSE AGREEMENT DATED JULY 14, 1999" (See file LICENSE.tt).
//;-------------------------------------------------------------------------

/******************************************************************************

  PicoWeb Parallel Port Programming Cable

  PC PP  Breadboard    PicoWeb                
  DB-25  8-Pin Header  DB-25    Signal Name
  -----  ------------  -------  -----------
    2       JP1-1        20     RESET* (active low)
    3       JP1-3        18     SCK (PB7)
    4       JP1-7        14     MOSI (PB5)
   13       JP1-5        16     MISO (PB6)
   15       JP1-2         -     GND
   24       JP1-4         -     GND
   23       JP1-6         -     GND
   22       JP1-8         1     GND
   12         -           -     PaperOut      \ Loop-back (cable detect)
   14         -           -     AutoLineFeed* / Loop-back (cable detect)

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(WIN32)
#include <windows.h>
#include "SetPrivIoXP.h"
#include "PrintfPrivIoXP.h"
#endif

#include "ppio.h"

static unsigned char mosi = 0;
static int RESETHI  = 1;
static int RESETLOW = 0;
static int ecp = 0;
static int ecp_found = 0;

// standard PC LPT ports
static unsigned int lpt_base[3] = {0x03bc, 0x0378, 0x0278};
#define NUM_LPT_PORTS  (sizeof(lpt_base) / sizeof(lpt_base[0]))

// LPT data port: 8 DATA BITS
#define LPT_DATA 0

// LPT control port: X X DIR IRQ4EN *SEL INIT *AUTOFEED *STROBE (POR XX001011)
#define LPT_CONTROL 2
#define LPT_DIR_BIT (1 << 5)   /* data port direction (0=output) */
#define LPT_IRQ_BIT (1 << 4)   /* IRQ enable bit (via "ACK" input) */
#define LPT_PO_BIT  (1 << 1)   /* "Paper Out" output bit (a.k.a. "AutoLF") */
#define LPT_ALF_BIT (1 << 5)   /* "Auto Linefeed" input bit (inverted) */

// LPT status port: *BUSY *ACK P.OUT SEL *ERR X X X
#define LPT_STATUS 1

// LPT Extended Control Register (ECR) port: S2 S1 S0 X X X X X
#define LPT_ECR 0x402
#define LPT_ECR_MASK (~(7 << 5))   /* SPP mode has upper 3 bits zero */
#define LPT_ECR_STD_MODE ((0) << 5)   /* SPP "standard mode" */
#define LPT_ECR_BYT_MODE ((1) << 5)   /* SPP "byte mode" (reverse data) */
#define LPT_ECR_EPP_MODE ((4) << 5)   /* SPP "EPP mode" */

static int lptN = -1;
unsigned int _pp_io_base = 0;
int directIO = 0;
int pp_invert_ios = 0;

static long lpt_port_hits_per_msec = 2000L;    /* 1.5-2.0 usec per _pp_inp() */

unsigned int pp_get_port(void)   // return LPT I/O port (0 = none)
{
    return _pp_io_base;
}

int pp_get_lpt(void)   // return LPT number (-1=none, n=LPTn)
{
    return lptN;
}

void pp_set_hits_per_msec(long hits_per_msec)
{
    lpt_port_hits_per_msec = hits_per_msec;
}

long pp_get_hits_per_msec(void)
{
    return lpt_port_hits_per_msec;
}

int pp_delay_ms(int msecs)
{
    int data;
    register long count;

    count = lpt_port_hits_per_msec * msecs;
    while (count-- > 0L) {
        data = _pp_inp(_pp_io_base + LPT_CONTROL);
    }
    return data;
}

void pp_set_invert_ios(void)
{
    pp_invert_ios = 1;
}

static int pp_setup_io(
    int verbose,
    FILE *fd,
    int port,   // -1=any
    int *pPort,
    int *pBase
){
    static char ok = 0;
#if defined(WIN32)
#ifdef USE_ACCESS_IO_PORTS
    int AccessIoPorts(void);
#endif

    if (directIO) ok = 1;
#endif

    if (pPort) *pPort = port;

    if (ok) return 1;    // only do this once (if successful)

#if defined(WIN32)
#ifdef USE_ACCESS_IO_PORTS
    if (AccessIoPorts() != 0)  { // get permission for user-mode I/O
        fprintf(fd,
            "Need \"Act as part of the operating system\" privilege for direct I/O!\n");
        return 0;
    }
#else /* !USE_ACCESS_IO_PORTS */
    // get permission for user-mode direct I/O
    if (bIsWindowsNT()) {
        if (!SetPrivIoXP((verbose>1)?TRUE:FALSE, "", TRUE,
                          (port < 0) ? 0 : port, pPort, NULL, NULL,
                                                              NULL, 0)) {
            if (verbose) {
                fprintf(fd,
                    "Failed to get direct I/O privileges for port %d!\n", port);
            }
            return 0;
        }
    }
#endif /* !USE_ACCESS_IO_PORTS */
#endif

#ifdef LINUX
    if (iopl(3)) {     // get permission for user-mode I/O (must be root!)
        fprintf(fd, "Must be root for direct I/O!\n");
        if (fd == stderr) perror("iopl");
        return 0;
    }
#endif
    ok = 1;
    return 1;
}

static int io_check(int port)
{
#if defined(WIN32)
    _try {
        _pp_inp(port);  // do "test" I/O read
        return 1;   // all OK
    } _except( EXCEPTION_EXECUTE_HANDLER ) {
        return 0;   // exception...return error
    }
#else
    return 1;
#endif
}

int pp_set_lpt(      // returned LPT number (-1=not found; n=LPTn)
    FILE *fd,        // log errors and messages here
    int verbose,     // 0=only log errors; 1=log progress reports too, 2=debug
    int ecp_mode,    // PC parallel port type: 0:legacy, 1:ECP
    int port         // <0: search for cable; >0: check LPTn for cable
){
    int ix, x;
    int first_port = 1;
    int last_port = NUM_LPT_PORTS;
    int i;
    int loop;
    int data;
    int base;
    int ecr;

    lptN = -1;         // default to no LPT port
    _pp_io_base = 0;

    // get any needed I/O privileges
    if (!pp_setup_io(verbose, fd, port, &port, &base)) {
        fflush(fd);
        return -1;
    } else {
        if (port >= 0 && port < NUM_LPT_PORTS) {
            lpt_base[port] = base;
        }
    }

    ecp = ecp_mode;
    if (port >= 0) {
        first_port = last_port = port;
    }
    // loop thru all possible parallel ports looking for loop-back signal
    for (ix = first_port; ix <= last_port; ix++) {
        x = ix % NUM_LPT_PORTS;
        ecp_found = 0;
        if (!io_check(lpt_base[x] + LPT_CONTROL)) {
            if (verbose > 1)
                fprintf(fd, "LPT%d (0x%x): direct I/O failed!\n",
                        x, lpt_base[x]),fflush(stdout);
            continue;
        }
        if (ecp) {
            ecr = _pp_inp(lpt_base[x] + LPT_ECR);
            if (verbose > 1)
                fprintf(fd, "LPT%d (0x%x): ECR=0x%02x\n",
                    x, lpt_base[x], ecr),fflush(stdout);
            _pp_outp((lpt_base[x] + LPT_ECR), (ecr & LPT_ECR_MASK) | LPT_ECR_BYT_MODE);
            i = _pp_inp(lpt_base[x] + LPT_ECR);
            if (i == ((ecr & LPT_ECR_MASK) | LPT_ECR_BYT_MODE)) {
                ecp_found = 1;
            }
            if (verbose > 1)
                fprintf(fd, "\tECR=0x%02x%s\n",
                    i, (ecp_found)?" [ECP]":""),fflush(stdout);
        }
        data = _pp_inp(lpt_base[x] + LPT_CONTROL);  // get/save port state
        _pp_outp((lpt_base[x] + LPT_CONTROL), (data & (~LPT_IRQ_BIT)) | LPT_PO_BIT | LPT_DIR_BIT);
        for (loop = 0; loop < 1000; ++loop) // delay for slow ports!!!??
            i = _pp_inp(lpt_base[x] + LPT_STATUS);
        if (verbose > 1)
            fprintf(fd, "LPT%d (0x%x): status=0x%02x\n",
                x, lpt_base[x], i),fflush(stdout);
        if ((i & LPT_ALF_BIT) != 0)  // "paper out" bit found?
            goto skip;
        _pp_outp((lpt_base[x] + LPT_CONTROL), (data & (~(LPT_PO_BIT | LPT_IRQ_BIT))) | LPT_DIR_BIT);
        for (loop = 0; loop < 1000; ++loop) // delay for slow ports!!!??
            i = _pp_inp(lpt_base[x] + LPT_STATUS);
        if (verbose > 1)
            fprintf(fd, "\tstatus=0x%02x\n", i),fflush(stdout);
        if ((i & LPT_ALF_BIT) != LPT_ALF_BIT)  // "paper out" bit found?
            goto skip;
        if (verbose) {
            fprintf(fd, "Found %sLPT%d (I/O base 0x%x)\n",
                    ecp_found ? "enhanced " : "", x, lpt_base[x]);
        }
        // activate reset
        _pp_io_base = lpt_base[x];
        pp_write(0, 0, RESETLOW);
        // set ECP to "standard mode" if ECP found
        if (ecp_found) {
            ecr = _pp_inp(lpt_base[x] + LPT_ECR);
            _pp_outp((lpt_base[x] + LPT_ECR), (ecr & LPT_ECR_MASK) | LPT_ECR_STD_MODE);
        }
        // set LPT data bus to output using control port's "direction" bit
        // with IRQ inactive
        data = _pp_inp(lpt_base[x] + LPT_CONTROL);
        _pp_outp((lpt_base[x] + LPT_CONTROL), (data | LPT_PO_BIT) & (~(LPT_DIR_BIT | LPT_IRQ_BIT)));
        // re-activate reset
        pp_write(0, 0, RESETLOW);
        lptN = x;
        fflush(fd);
        return lptN;
      skip:
        if (verbose > 1) {
            i = _pp_inp(lpt_base[x] + LPT_STATUS);
            fprintf(fd, "\tstatus=0x%02x\n", i),fflush(stdout);
        }
        _pp_outp((lpt_base[x] + LPT_CONTROL), data);  // restore port state
        if (ecp) {
            _pp_outp((lpt_base[x] + LPT_ECR), ecr);  // restore ECR port state
        }
    }
    if (port >= 0) {
        fprintf(fd, "Can't find the AVR program cable on LPT%d!\n", port);
    } else {
        fprintf(fd, "Can't find the AVR program cable on any LPT port!\n");
    }
    fflush(fd);
    return -1;
}

void pp_reset(void)
{
    if (_pp_io_base != 0) {
        /* delay about 1 msec with RESET active */
        pp_write(0, 0, RESETLOW);
	pp_delay_ms(1);
        pp_write(0, 0, RESETHI);
    }
}

void pp_cleanup(
    int reset_active,  // 1=reset active, 0=reset inactive, -1=don't touch
    int bus_active     // leave parallel port output drivers active
){
    int i;

    if (_pp_io_base != 0) {
        if (reset_active >= 0) {
            if (reset_active) {
                pp_write(0, 0, RESETLOW);
            } else {
                pp_write(0, 0, RESETHI);
            }
        }
        if (!bus_active) {
            // set ECP to "byte mode" (reverses data bus)
            if (ecp_found) {
                i = _pp_inp(_pp_io_base + LPT_ECR);
                _pp_outp(_pp_io_base + LPT_ECR, (i & LPT_ECR_MASK) | LPT_ECR_BYT_MODE);
            }
            // set LPT data direction to input (IRQ inactive)
            i = _pp_inp(_pp_io_base + LPT_CONTROL);  // get control port state
            _pp_outp(_pp_io_base + LPT_CONTROL, (i & (~LPT_IRQ_BIT)) | LPT_DIR_BIT);
        }
    }
#ifdef LINUX
    iopl(0);   // release our special I/O permissions
#endif
    _pp_io_base = 0;
    lptN = -1;
}

